/* -*- mode: C++; c-basic-offset: 4; tab-width: 4 -*-
 *
 * Copyright (c) 2008 Apple Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 *
 * This file contains Original Code and/or Modifications of Original Code
 * as defined in and that are subject to the Apple Public Source License
 * Version 2.0 (the 'License'). You may not use this file except in
 * compliance with the License. Please obtain a copy of the License at
 * http://www.opensource.apple.com/apsl/ and read it before using this
 * file.
 *
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 * Please see the License for the specific language governing rights and
 * limitations under the License.
 *
 * @APPLE_LICENSE_HEADER_END@
 */
 
//
//	C++ interface to lower levels of libuwind 
//

#ifndef __ADDRESSSPACE_HPP__
#define __ADDRESSSPACE_HPP__

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>
#include <mach-o/loader.h>
#include <mach-o/getsect.h>
//#include <mach-o/dyld_priv.h>

// from http://www.opensource.apple.com/source/dyld/dyld-353.2.1/include/mach-o/dyld_priv.h
/* -*- mode: C++; c-basic-offset: 4; tab-width: 4 -*-
 *
 * Copyright (c) 2003-2010 Apple Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 *
 * This file contains Original Code and/or Modifications of Original Code
 * as defined in and that are subject to the Apple Public Source License
 * Version 2.0 (the 'License'). You may not use this file except in
 * compliance with the License. Please obtain a copy of the License at
 * http://www.opensource.apple.com/apsl/ and read it before using this
 * file.
 *
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 * Please see the License for the specific language governing rights and
 * limitations under the License.
 *
 * @APPLE_LICENSE_HEADER_END@
 */
#ifndef _MACH_O_DYLD_PRIV_H_
#define _MACH_O_DYLD_PRIV_H_

#include <stdbool.h>
#include <Availability.h>
#include <mach-o/dyld.h>
#include <mach-o/dyld_images.h>

#if __cplusplus
extern "C" {
#endif /* __cplusplus */
    
    
    
    //
    // private interface between libSystem.dylib and dyld
    //
    extern int _dyld_func_lookup(const char* dyld_func_name, void **address);
    
    //
    // private interface between libSystem.dylib and dyld
    //
    extern void _dyld_fork_child();
    
    
    //
    // Possible state changes for which you can register to be notified
    //
    enum dyld_image_states
    {
        dyld_image_state_mapped					= 10,		// No batch notification for this
        dyld_image_state_dependents_mapped		= 20,		// Only batch notification for this
        dyld_image_state_rebased				= 30,
        dyld_image_state_bound					= 40,
        dyld_image_state_dependents_initialized	= 45,		// Only single notification for this
        dyld_image_state_initialized			= 50,
        dyld_image_state_terminated				= 60		// Only single notification for this
    };
    
    //
    // Callback that provides a bottom-up array of images
    // For dyld_image_state_[dependents_]mapped state only, returning non-NULL will cause dyld to abort loading all those images
    // and append the returned string to its load failure error message. dyld does not free the string, so
    // it should be a literal string or a static buffer
    //
    typedef const char* (*dyld_image_state_change_handler)(enum dyld_image_states state, uint32_t infoCount, const struct dyld_image_info info[]);
    
    //
    // Register a handler to be called when any image changes to the requested state.
    // If 'batch' is true, the callback is called with an array of all images that are in the requested state sorted by dependency.
    // If 'batch' is false, the callback is called with one image at a time as each image transitions to the the requested state.
    // During the call to this function, the handler may be called back with existing images and the handler should
    // not return a string, since there is no load to abort.  In batch mode, existing images at or past the request
    // state supplied in the callback.  In non-batch mode, the callback is called for each image exactly in the
    // requested state.
    //
    extern void
    dyld_register_image_state_change_handler(enum dyld_image_states state, bool batch, dyld_image_state_change_handler handler);
    
    
    //
    // Possible thread-local variable state changes for which you can register to be notified
    //
    enum dyld_tlv_states {
        dyld_tlv_state_allocated = 10,   // TLV range newly allocated
        dyld_tlv_state_deallocated = 20  // TLV range about to be deallocated
    };
    
    //
    // Info about thread-local variable storage.
    //
    typedef struct {
        size_t info_size;    // sizeof(dyld_tlv_info)
        void * tlv_addr;     // Base address of TLV storage
        size_t tlv_size;     // Byte size of TLV storage
    } dyld_tlv_info;
    
#if __BLOCKS__
    
    //
    // Callback that notes changes to thread-local variable storage.
    //
    typedef void (^dyld_tlv_state_change_handler)(enum dyld_tlv_states state, const dyld_tlv_info *info);
    
    //
    // Register a handler to be called when a thread adds or removes storage for thread-local variables.
    // The registered handler will only be called from and on behalf of the thread that owns the storage.
    // The registered handler will NOT be called for any storage that was
    //   already allocated before dyld_register_tlv_state_change_handler() was
    //   called. Use dyld_enumerate_tlv_storage() to get that information.
    // Exists in Mac OS X 10.7 and later
    //
    extern void
    dyld_register_tlv_state_change_handler(enum dyld_tlv_states state, dyld_tlv_state_change_handler handler);
    
    //
    // Enumerate the current thread-local variable storage allocated for the current thread.
    // Exists in Mac OS X 10.7 and later
    //
    extern void
    dyld_enumerate_tlv_storage(dyld_tlv_state_change_handler handler);
    
#endif
    
    
    //
    // get slide for a given loaded mach_header
    // Mac OS X 10.6 and later
    //
    extern intptr_t _dyld_get_image_slide(const struct mach_header* mh);
    
    
    //
    // get pointer to this process's dyld_all_image_infos
    // Exists in Mac OS X 10.4 and later through _dyld_func_lookup()
    // Exists in Mac OS X 10.6 and later through libSystem.dylib
    //
    const struct dyld_all_image_infos* _dyld_get_all_image_infos();
    
    
    
    struct dyld_unwind_sections
    {
        const struct mach_header*		mh;
        const void*						dwarf_section;
        uintptr_t						dwarf_section_length;
        const void*						compact_unwind_section;
        uintptr_t						compact_unwind_section_length;
    };
    
    
    //
    // Returns true iff some loaded mach-o image contains "addr".
    //	info->mh							mach header of image containing addr
    //  info->dwarf_section					pointer to start of __TEXT/__eh_frame section
    //  info->dwarf_section_length			length of __TEXT/__eh_frame section
    //  info->compact_unwind_section		pointer to start of __TEXT/__unwind_info section
    //  info->compact_unwind_section_length	length of __TEXT/__unwind_info section
    //
    // Exists in Mac OS X 10.6 and later
    extern bool _dyld_find_unwind_sections(void* addr, struct dyld_unwind_sections* info);
    
    
    //
    // This is an optimized form of dladdr() that only returns the dli_fname field.
    //
    // Exists in Mac OS X 10.6 and later
    extern const char* dyld_image_path_containing_address(const void* addr);
    
    
    
    // Convienence constants for return values from dyld_get_sdk_version() and friends.
#define DYLD_MACOSX_VERSION_10_4		0x000A0400
#define DYLD_MACOSX_VERSION_10_5		0x000A0500
#define DYLD_MACOSX_VERSION_10_6		0x000A0600
#define DYLD_MACOSX_VERSION_10_7		0x000A0700
#define DYLD_MACOSX_VERSION_10_8		0x000A0800
#define DYLD_MACOSX_VERSION_10_9		0x000A0900
#define DYLD_MACOSX_VERSION_10_10		0x000A0A00
    
#define DYLD_IOS_VERSION_2_0		0x00020000
#define DYLD_IOS_VERSION_2_1		0x00020100
#define DYLD_IOS_VERSION_2_2		0x00020200
#define DYLD_IOS_VERSION_3_0		0x00030000
#define DYLD_IOS_VERSION_3_1		0x00030100
#define DYLD_IOS_VERSION_3_2		0x00030200
#define DYLD_IOS_VERSION_4_0		0x00040000
#define DYLD_IOS_VERSION_4_1		0x00040100
#define DYLD_IOS_VERSION_4_2		0x00040200
#define DYLD_IOS_VERSION_4_3		0x00040300
#define DYLD_IOS_VERSION_5_0		0x00050000
#define DYLD_IOS_VERSION_5_1		0x00050100
#define DYLD_IOS_VERSION_6_0		0x00060000
#define DYLD_IOS_VERSION_6_1		0x00060100
#define DYLD_IOS_VERSION_7_0		0x00070000
#define DYLD_IOS_VERSION_7_1		0x00070100
#define DYLD_IOS_VERSION_8_0		0x00080000
    
    //
    // This is finds the SDK version a binary was built against.
    // Returns zero on error, or if SDK version could not be determined.
    //
    // Exists in Mac OS X 10.8 and later
    // Exists in iOS 6.0 and later
    extern uint32_t dyld_get_sdk_version(const struct mach_header* mh);
    
    
    //
    // This is finds the SDK version the main executable was built against.
    // Returns zero on error, or if SDK version could not be determined.
    //
    // Exists in Mac OS X 10.8 and later
    // Exists in iOS 6.0 and later
    extern uint32_t dyld_get_program_sdk_version();
    
    
    //
    // This is finds the min OS version a binary was built to run on.
    // Returns zero on error, or if no min OS recorded in binary.
    //
    // Exists in Mac OS X 10.8 and later
    // Exists in iOS 6.0 and later
    extern uint32_t dyld_get_min_os_version(const struct mach_header* mh);
    
    
    //
    // This is finds the min OS version the main executable was built to run on.
    // Returns zero on error, or if no min OS recorded in binary.
    //
    // Exists in Mac OS X 10.8 and later
    // Exists in iOS 6.0 and later
    extern uint32_t dyld_get_program_min_os_version();
    
    //
    // Returns if any OS dylib has overridden its copy in the shared cache
    //
    // Exists in iPhoneOS 3.1 and later
    // Exists in Mac OS X 10.10 and later
    extern bool dyld_shared_cache_some_image_overridden();
	
    //
    // Returns if the process is setuid or is code signed with entitlements.
    //
    // Exists in Mac OS X 10.9 and later
    extern bool dyld_process_is_restricted();
    
    
    //
    // <rdar://problem/13820686> for OpenGL to tell dyld it is ok to deallocate a memory based image when done.
    //
    // Exists in Mac OS X 10.9 and later
#define NSLINKMODULE_OPTION_CAN_UNLOAD                  0x20
    
    
    //
    // Update all bindings on specified image. 
    // Looks for uses of 'replacement' and changes it to 'replacee'.
    // NOTE: this is less safe than using static interposing via DYLD_INSERT_LIBRARIES
    // because the running program may have already copy the pointer values to other
    // locations that dyld does not know about.
    //
    struct dyld_interpose_tuple {
        const void* replacement;
        const void* replacee;
    };
    extern void dyld_dynamic_interpose(const struct mach_header* mh, const struct dyld_interpose_tuple array[], size_t count);
    
    
    
#if __cplusplus
}
#endif /* __cplusplus */

#endif /* _MACH_O_DYLD_PRIV_H_ */



#include <mach/i386/thread_status.h>
#include <Availability.h>

#include "FileAbstraction.hpp"
#include "libunwind.h"
#include "InternalMacros.h"
#include "dwarf2.h"


#if 0
#if __i386__ || __x86_64__ 
// In 10.6 and later i386 and x86_64 don't have a __dyld section
// We need one to access private _dyld_func_lookup function.

struct __DATA__dyld { long lazy; int (*lookup)(const char*, void**); };

static volatile struct __DATA__dyld  myDyldSection __attribute__ ((section ("__DATA,__dyld"))) = { 0, NULL };


static int my_dyld_func_lookup(const char* dyld_func_name, void **address)
{
	return (*myDyldSection.lookup)(dyld_func_name, address);
}
#else
	#define my_dyld_func_lookup _dyld_func_lookup
#endif


bool _dyld_find_unwind_sections(void* addr, dyld_unwind_sections* info)
{
    static void* (*p)(void*, dyld_unwind_sections*) = NULL;

	if(p == NULL)
	    my_dyld_func_lookup("__dyld_find_unwind_sections", (void**)&p);
	return p(addr, info);
}
#endif // 0



namespace libunwind {

///
/// LocalAddressSpace is used as a template parameter to UnwindCursor when unwinding a thread
/// in the same process.  It compiles away and making local unwinds very fast.
///
class LocalAddressSpace
{
public:
	
	#if __LP64__
		typedef uint64_t	pint_t;
		typedef  int64_t	sint_t;
	#else
		typedef uint32_t	pint_t;
		typedef  int32_t	sint_t;
	#endif
		uint8_t			get8(pint_t addr)	{ return *((uint8_t*)addr); }
		uint16_t		get16(pint_t addr)	{ return *((uint16_t*)addr); }
		uint32_t		get32(pint_t addr)	{ return *((uint32_t*)addr); }
		uint64_t		get64(pint_t addr)	{ return *((uint64_t*)addr); }
		double			getDouble(pint_t addr)	{ return *((double*)addr); }
		v128			getVector(pint_t addr)	{ return *((v128*)addr); }
		uintptr_t		getP(pint_t addr);
	static uint64_t		getULEB128(pint_t& addr, pint_t end);
	static int64_t		getSLEB128(pint_t& addr, pint_t end);
	
		pint_t			getEncodedP(pint_t& addr, pint_t end, uint8_t encoding);
		bool			findFunctionName(pint_t addr, char* buf, size_t bufLen, unw_word_t* offset);
		bool			findUnwindSections(pint_t addr, pint_t& mh, pint_t& dwarfStart, pint_t& dwarfLen, pint_t& compactStart);
	
};

LocalAddressSpace sThisAddress;

inline uintptr_t LocalAddressSpace::getP(pint_t addr)
{
#if __LP64__
	return get64(addr);
#else
	return get32(addr);
#endif
}

/* Read a ULEB128 into a 64-bit word.   */
inline uint64_t
LocalAddressSpace::getULEB128(pint_t& addr, pint_t end)
{
	const uint8_t* p = (uint8_t*)addr;
	const uint8_t* pend = (uint8_t*)end;
	uint64_t result = 0;
	int bit = 0;
	do  {
		uint64_t b;

		if ( p == pend )
			ABORT("truncated uleb128 expression");

		b = *p & 0x7f;

		if (bit >= 64 || b << bit >> bit != b) {
			ABORT("malformed uleb128 expression");
		}
		else {
			result |= b << bit;
			bit += 7;
		}
	} while ( *p++ >= 0x80 );
	addr = (pint_t)p;
	return result;
}

/* Read a SLEB128 into a 64-bit word.  */
inline int64_t
LocalAddressSpace::getSLEB128(pint_t& addr, pint_t end)
{
	const uint8_t* p = (uint8_t*)addr;
	int64_t result = 0;
	int bit = 0;
	uint8_t byte;
	do {
		byte = *p++;
		result |= ((byte & 0x7f) << bit);
		bit += 7;
	} while (byte & 0x80);
	// sign extend negative numbers
	if ( (byte & 0x40) != 0 )
		result |= (-1LL) << bit;
	addr = (pint_t)p;
	return result;
}

LocalAddressSpace::pint_t
LocalAddressSpace::getEncodedP(pint_t& addr, pint_t end, uint8_t encoding)
{
	pint_t startAddr = addr;
	const uint8_t* p = (uint8_t*)addr;
	pint_t result;
	
	// first get value
	switch (encoding & 0x0F) {
		case DW_EH_PE_ptr:
			result = getP(addr);
			p += sizeof(pint_t);
			addr = (pint_t)p;
			break;
		case DW_EH_PE_uleb128:
			result = getULEB128(addr, end);
			break;
		case DW_EH_PE_udata2:
			result = get16(addr);
			p += 2;
			addr = (pint_t)p;
			break;
		case DW_EH_PE_udata4:
			result = get32(addr);
			p += 4;
			addr = (pint_t)p;
			break;
		case DW_EH_PE_udata8:
			result = get64(addr);
			p += 8;
			addr = (pint_t)p;
			break;
		case DW_EH_PE_sleb128:
			result = getSLEB128(addr, end);
			break;
		case DW_EH_PE_sdata2:
			result = (int16_t)get16(addr);
			p += 2;
			addr = (pint_t)p;
			break;
		case DW_EH_PE_sdata4:
			result = (int32_t)get32(addr);
			p += 4;
			addr = (pint_t)p;
			break;
		case DW_EH_PE_sdata8:
			result = get64(addr);
			p += 8;
			addr = (pint_t)p;
			break;
		default:
			ABORT("unknown pointer encoding");
	}
	
	// then add relative offset
	switch ( encoding & 0x70 ) {
		case DW_EH_PE_absptr:
			// do nothing
			break;
		case DW_EH_PE_pcrel:
			result += startAddr;
			break;
		case DW_EH_PE_textrel:
			ABORT("DW_EH_PE_textrel pointer encoding not supported");
			break;
		case DW_EH_PE_datarel:
			ABORT("DW_EH_PE_datarel pointer encoding not supported");
			break;
		case DW_EH_PE_funcrel:
			ABORT("DW_EH_PE_funcrel pointer encoding not supported");
			break;
		case DW_EH_PE_aligned:
			ABORT("DW_EH_PE_aligned pointer encoding not supported");
			break;
		default:
			ABORT("unknown pointer encoding");
			break;
	}
	
	if ( encoding & DW_EH_PE_indirect )
		result = getP(result);
	
	return result;
}


inline bool LocalAddressSpace::findUnwindSections(pint_t addr, pint_t& mh, pint_t& dwarfStart, pint_t& dwarfLen, pint_t& compactStart)
{
	dyld_unwind_sections info;
	if ( _dyld_find_unwind_sections((void*)addr, &info) ) {
		mh				= (pint_t)info.mh;
		dwarfStart		= (pint_t)info.dwarf_section;
		dwarfLen		= (pint_t)info.dwarf_section_length;
		compactStart	= (pint_t)info.compact_unwind_section;
		return true;
	}
	return false;
}


inline bool	LocalAddressSpace::findFunctionName(pint_t addr, char* buf, size_t bufLen, unw_word_t* offset)
{
	dl_info dyldInfo;
	if ( dladdr((void*)addr, &dyldInfo) ) {
		if ( dyldInfo.dli_sname != NULL ) {
			strlcpy(buf, dyldInfo.dli_sname, bufLen);
			*offset = (addr - (pint_t)dyldInfo.dli_saddr);
			return true;
		}
	}
	return false;
}



#if UNW_REMOTE

///
/// OtherAddressSpace is used as a template parameter to UnwindCursor when unwinding a thread
/// in the another process.  The other process can be a different endianness and a different
/// pointer size and is handled by the P template parameter.  
///
template <typename P>
class OtherAddressSpace
{
public:
						OtherAddressSpace(task_t task) : fTask(task) {}

		typedef typename P::uint_t	pint_t;

		uint8_t			get8(pint_t addr);
		uint16_t		get16(pint_t addr);
		uint32_t		get32(pint_t addr);
		uint64_t		get64(pint_t addr);
		pint_t			getP(pint_t addr);
		uint64_t		getULEB128(pint_t& addr, pint_t end);
		int64_t			getSLEB128(pint_t& addr, pint_t end);
		pint_t			getEncodedP(pint_t& addr, pint_t end, uint8_t encoding);
		bool			findFunctionName(pint_t addr, char* buf, size_t bufLen, unw_word_t* offset);
		bool			findUnwindSections(pint_t addr, unwind_sections& info);
private:
		void*			localCopy(pint_t addr);


		task_t			fTask;
};


template <typename P>
uint8_t OtherAddressSpace<P>::get8(pint_t addr)
{
	return *((uint8_t*)localCopy(addr));
}

template <typename P>
uint16_t OtherAddressSpace<P>::get16(pint_t addr)
{
	return P::E::get16(*(uint16_t*)localCopy(addr));
}

template <typename P>
uint32_t OtherAddressSpace<P>::get32(pint_t addr)
{
	return P::E::get32(*(uint32_t*)localCopy(addr));
}

template <typename P>
uint64_t OtherAddressSpace<P>::get64(pint_t addr)
{
	return P::E::get64(*(uint64_t*)localCopy(addr));
}

template <typename P>
typename P::uint_t OtherAddressSpace<P>::getP(pint_t addr)
{
	return P::getP(*(uint64_t*)localCopy(addr));
}

template <typename P>
uint64_t OtherAddressSpace<P>::getULEB128(pint_t& addr, pint_t end)
{
	uintptr_t size = (end - addr);
	LocalAddressSpace::pint_t laddr = (LocalAddressSpace::pint_t)localCopy(addr);
	LocalAddressSpace::pint_t sladdr = laddr;
	uint64_t result = LocalAddressSpace::getULEB128(laddr, laddr+size);
	addr += (laddr-sladdr);
	return result;
}

template <typename P>
int64_t OtherAddressSpace<P>::getSLEB128(pint_t& addr, pint_t end)
{
	uintptr_t size = (end - addr);
	LocalAddressSpace::pint_t laddr = (LocalAddressSpace::pint_t)localCopy(addr);
	LocalAddressSpace::pint_t sladdr = laddr;
	uint64_t result = LocalAddressSpace::getSLEB128(laddr, laddr+size);
	addr += (laddr-sladdr);
	return result;
}

template <typename P>
void* OtherAddressSpace<P>::localCopy(pint_t addr)
{
	// FIX ME
}

template <typename P>
bool OtherAddressSpace<P>::findFunctionName(pint_t addr, char* buf, size_t bufLen, unw_word_t* offset)
{
	// FIX ME
}



///
/// unw_addr_space is the base class that abstract unw_addr_space_t type in libunwind.h points to.  
///
struct unw_addr_space
{
	cpu_type_t				cpuType;
	task_t					taskPort;
};


///
/// unw_addr_space_i386 is the concrete instance that a unw_addr_space_t points to when examining
/// a 32-bit intel process.
///
struct unw_addr_space_i386 : public unw_addr_space
{
													unw_addr_space_i386(task_t task) : oas(task) {}
	OtherAddressSpace<Pointer32<LittleEndian> >		oas;
};


///
/// unw_addr_space_x86_64 is the concrete instance that a unw_addr_space_t points to when examining
/// a 64-bit intel process.
///
struct unw_addr_space_x86_64 : public unw_addr_space
{
													unw_addr_space_x86_64(task_t task) : oas(task) {}
	OtherAddressSpace<Pointer64<LittleEndian> >		oas;
};


///
/// unw_addr_space_ppc is the concrete instance that a unw_addr_space_t points to when examining
/// a 32-bit PowerPC process.
///
struct unw_addr_space_ppc : public unw_addr_space
{
													unw_addr_space_ppc(task_t task) : oas(task) {}
	OtherAddressSpace<Pointer32<BigEndian> >		oas;
};


#endif // UNW_REMOTE


} // namespace libunwind 



#endif // __ADDRESSSPACE_HPP__




